#   Copyright (c) 1997 Boulder Real Time Technologies, Inc.           
#                                                                     
#   This software module is wholly owned by Boulder Real Time         
#   Technologies, Inc. Any use of this software module without        
#   express written permission from Boulder Real Time Technologies,   
#   Inc. is prohibited.                                               

# Each task with an entry in the Run array in rtexec 
# has a series of gauges and labels in the process
# window.  

package provide Procwin 1.0

namespace eval Procwin { 

    variable Root	    ; # frame which contains buttons and activity bars
    variable Frame	    ; # frame which contains scrollbar and canvas
    variable Rtsys	    ; # db pointer for rtsys database
    variable RtsysFile      ; # file name for rtsys process table
    variable RtsysMtime	    ; # mtime for process table

    variable TaskButton	    ; # array of TaskButtons indexed by task name
    variable Activity	    ; # array indexed by task names of lists of activity widgets
    variable Dim	    ; # array of flags indexed by task name: ==1 if activity widgets are "dimmed"
    variable Reset	    ; # array of reset yellow flags
    variable Variable	    ; # array of variable names indexed by widget name

    variable Stat	    ; # name of measurement for a column, indexed by column name
    variable Gauges	    ; # array of lists of gauges, indexed by column name (eg, g1, g2..)
    variable Labels	    ; # array of lists of labels, indexed by column name (eg, l1, l2..)

    variable RtexecMtime    ; # mtime for rtexec.pf
    variable Run	    ; # array from rtexec.pf of tasks to run 
    variable Processes      ; # list from rtexec.pf of tasks and execution lines
    variable Cmdline	    ; # array of command lines indexed by task name
    variable Parameter_files ; # array from rtexec.pf of parameter files indexed by task name

    variable ProcStats	    ; # list from rtm.pf of process statistics
    variable ShowRtm	    ; # flag for whether to show rtmonitor (this process)
    variable OrbWidgets	    ; # array of orb widgets, indexed by task name
    variable OrbMeasure	    ; # array of orb measurements, indexed by widget
    variable HeaderHeight   ; # size header in process panel
    variable HasScrollbar   ; # name of scroll bar if present
    variable TaskColor	    ; # array of TaskColor by task name

# global T_$name	    statistics from statproc $pid for task $name
# global Pid		    array of pid indexed by task name

    proc gridrow {w} {
	set info [grid info $w]
	set n [lsearch $info -row]
	set row [lindex $info [incr n]]
	return $row
    }

    proc dialogMessage {w title msg variable} { 
	toplevel $w 

	label $w.bitmap -bitmap info -width 50 -height 50
	message $w.message \
	    -width 300 \
	    -text $msg \
	    -foreground red \
	    -font Emphasis
	pack $w.bitmap $w.message -side left -expand yes -fill both

	wm title $w $title
	wm iconname $w $title
	wm transient $w [winfo toplevel [winfo parent $w]]
	wm withdraw $w

	# update idletasks
	set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
		- [winfo vrootx [winfo parent $w]]]
	set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
		- [winfo vrooty [winfo parent $w]]]
	wm geom $w +$x+$y

	wm deiconify $w

	set oldFocus [focus]
	set oldGrab [grab current $w]
	if {$oldGrab != ""} {
	    set grabStatus [grab status $oldGrab]
	}
	grab $w
	focus $w

	global Pid
	if { [info exists $variable] && [set $variable] != "" } { 
	    tkwait variable $variable
	}

	catch {focus $oldFocus}
	catch {
	    # It's possible that the window has already been destroyed,
	    # hence this "catch".  Delete the Destroy handler so that
	    # tkPriv(button) doesn't get reset by it.
     
	    bind $w <Destroy> {}
	    destroy $w
	}
	if {$oldGrab != ""} {
	    if {$grabStatus == "global"} {
		grab -global $oldGrab
	    } else {
		grab $oldGrab
	    }
	}
    }

    proc wait_for_stop { name variable pid w } { 
	if { [pid_exists $pid] } { 
	    after 100 "Procwin::wait_for_stop $name $variable $pid $w"
	} else { 
	    catch {$w config -text "restarting $name"}
	    variable TaskButton
	    after 3000 "set $variable 0"
	}
    }

    proc createActivity {name} {
	variable Root

	set w $Root

	variable Stat

	global Pid
	if { ! [info exists Pid($name)] } { 
	    set Pid($name) -1
	}
	set p $w.p_$name
	label $p \
	    -font Small \
	    -width 5 \
	    -textvariable Pid($name)

	variable Variable

	set l1 $w.l1_$name
	variable Labels
	lappend Labels(l1) $l1
	label $l1 -textvariable T_${name}($Stat(l1)) -width 5
	set Variable($l1) T_$name

	set g1 $w.g1_$name
	variable Gauges
	lappend Gauges(g1) $g1
	gauge::create $g1 \
	    -textvariable T_${name}($Stat(g1)) \
	    -rangevar g1 \
	    -annotate yes
	set Variable($g1) T_$name

	set l2 $w.l2_$name
	variable Labels
	lappend Labels(l2) $l2
	label $l2 -textvariable T_${name}($Stat(l2)) -width 5
	set Variable($l2) T_$name

	set g2 $w.g2_$name
	variable Gauges
	lappend Gauges(g2) $g2
	gauge::create $g2 \
	    -textvariable T_${name}($Stat(g2)) \
	    -rangevar g2 \
	    -annotate yes
	set Variable($g2) T_$name

	variable TaskButton
	set row [gridrow $TaskButton($name)]

	set col 0
	grid $p  -row $row -column [incr col] -stick ens
	grid $l1 -row $row -column [incr col] -stick ens
	grid $g1 -row $row -column [incr col] -stick wns
	grid $l2 -row $row -column [incr col] -stick ens
	grid $g2 -row $row -column [incr col] -stick wns

	variable Activity
	set Activity($name) "$p $l1 $g1 $l2 $g2"
	set Gauges($name) "$g1 $g2"
	variable Dim
        set Dim($name) 0

	global Pid
	pfgetarr orbtasks @rtexec#orbtasks 
	global Clients$Pid($name)
	variable OrbWidgets
	if { [info exists orbtasks($name)] } { 
	    variable OrbMeasure

	    set g3 $w.g3_$name
	    set OrbMeasure($g3) ipktrate
	    lappend Gauges(g3) $g3
	    gauge::create $g3 \
		-textvariable Clients$Pid($name)(ipktrate) \
		-autorange yes \
		-colorlist steelblue \
		-rangevar g3 \
		-annotate yes

	    set g4 $w.g4_$name
	    lappend Gauges(g4) $g4
	    set OrbMeasure($g4) opktrate
	    gauge::create $g4 \
		-textvariable Clients$Pid($name)(opktrate) \
		-autorange yes \
		-colorlist steelblue \
		-rangevar g4 \
		-annotate yes

	    set g5 $w.g5_$name
	    lappend Gauges(g5) $g5
	    set OrbMeasure($g5) latency
	    gauge::create $g5 \
		-textvariable Clients$Pid($name)(latency) \
		-autorange yes \
		-colorlist steelblue \
		-rangevar g5 \
		-annotate yes

	    grid $g3 -row $row -column [incr col] -stick wns
	    grid $g4 -row $row -column [incr col] -stick wns
	    grid $g5 -row $row -column [incr col] -stick wns
	    append Activity($name) " $g3 $g4 $g5"
	    append Gauges($name) " $g3 $g4 $g5"
	    set OrbWidgets($name) "$g3 $g4 $g5"
	} else { 
	    set OrbWidgets($name) {}
	}
    }

    proc dimActivity {name} { 
	variable Activity
	foreach w $Activity($name) {
	    $w configure -bg orange
	}
	global T_$name
	set T_${name}(cpu) 0
	set T_${name}(mem) 0
    }

    proc normalActivity {name} { 
	global Color
	variable Activity
	foreach w $Activity($name) {
	    $w configure -bg $Color(background)
	}
    }

    proc destroyActivity {name} {
	variable Activity
	global Pid
	set Pid($name) {}
	variable Gauges
	if { [info exists Gauges($name)] } { 
	    foreach gauge $Gauges($name) {
		gauge::destroy $gauge 
	    }
	    unset -nocomplain Gauges($name)
	}
	if { [info exists Activity($name)] } { 
	    foreach w $Activity($name) {
		destroy $w
	    }
	    unset -nocomplain Activity($name) 
	}
    }

    proc destroyRow {name} { 
# eliminate one row of process window corresponding to task name
	variable TaskButton
	destroyActivity $name
	destroy $TaskButton($name)
	unset -nocomplain TaskButton($name)
    }

    proc adjust_tasklines {} {
# add or delete task buttons as necessary, and ajust the stop/start
# status appropriately
	global Color

	variable Processes
	variable TaskButton
	variable Activity
	variable Run

	variable Root
	set w $Root

	variable Cmdline
	catch {unset -nocomplain Cmdline}
	variable Cmdline
	foreach proc $Processes { 
	    set name [lkey $proc]
	    set cmdline [lval $proc]
	    set Cmdline($name) $cmdline

	    if { [info exists Run($name)] } {
		if { [info exists TaskButton($name)] } {
		    # already have button, what about line?
		    if { $Run($name) && ! [info exists Activity($name)] } { 
			# Run changed to 1
			createActivity $name 
		    } elseif { ! $Run($name) && [info exists Activity($name)] } {
			# Run changed to 0
			global Pid
			unset -nocomplain Pid($name)
			$TaskButton($name) config -bg $Color(background)
			destroyActivity $name
		    }
		    setTaskButton $name
		} else { 
		    taskButton $w.task_$name $Run($name) $name $cmdline
		    if { $Run($name) } {
			createActivity $name 
		    }
		}
		if [ info exists Cmdline($name) ] {
		    set_balloon $TaskButton($name) $Cmdline($name)
		}
	    }
	}
	foreach name [array names TaskButton] { 
	    if { ! [info exists Cmdline($name)] 
		  || ! [info exists Run($name)] } { 
		destroyRow $name
	    }
	}
    }

    proc fix_Run {} {
	variable Run
	set bad {}
	foreach i [array names Run] { 
	    set flag [yesno $Run($i)]
	    switch -- $flag {
		maybe { 
		    append bad "  Run{$i} => '$Run($i)'\n" 
		    set Run($i) 0 
		}
		yes {set Run($i) 1 }
		no  {set Run($i) 0 }
	    }
	}
	if { $bad != "" } {
	    # after 10 "Procwin::dialogMessage .warning Warning \"$bad\" ''"
	    set bad "Invalid values in rtexec.pf:Run\n$bad"
	    after 10 "tk_dialog .warning Warning \"$bad\" error 0 ok"

	}
    }

    proc check_rtexec {} {
# reread rtexec.pf if necessary, and fill in variables appropriately
	set mtime [file mtime rtexec.pf]
	variable RtexecMtime
	if { $mtime != $RtexecMtime } {
	    pfread rtexec
	    set RtexecMtime $mtime

	    set process_database [pfget rtexec Database]
	    variable Rtsys
	    set Rtsys [dbopen $process_database r]
	    set Rtsys [dblookup $Rtsys "" process "" ""]
	    variable RtsysFile
	    set RtsysFile [dbquery $Rtsys dbTABLE_FILENAME]
	    variable RtsysMtime
	    set RtsysMtime 0

# add rtexec to process list explicitly, since it's not in rtexec's
# Run list
	    variable Processes
	    catch {unset -nocomplain Processes}
	    set Processes [pfgetlist @rtexec#Processes]
	    set Processes [concat {{rtexec rtexec}} $Processes]

	    variable Parameter_files
	    pfgetarr Parameter_files @rtexec#Parameter_files

	    variable Run
	    catch {unset -nocomplain Run}
	    pfgetarr Run @rtexec#Run
	    fix_Run
	    set Run(rtexec) 1
	    variable OrbWidgets
	    set OrbWidgets(rtexec) {}

# if rtm is not in rtexec's Run list, add it
	    variable ShowRtm
	    if { $ShowRtm } { 
		if { ! [info exists Run(rtm)] } {
		    set Processes [concat $Processes {{rtm rtm}}]
		    set Run(rtm) 1
		    global Pid
		    set Pid(rtm) [pid]
		    set OrbWidgets(rtm) {}
		}
	    }

	    adjust_tasklines
	}
    }

    proc taskcolor {name} { 
	global Pid Started Starts Now Color YellowTime
	global ResetTime
	variable TaskColor
	set TaskColor($name) $Color(background)
	if { [info exists Starts($name)]
	    && [info exists Started($name)]
	    && ($Starts($name) > 1) } { 
	    if {[expr $Now-$Started($name)] < $YellowTime 
		&& [expr $Started($name) - $ResetTime($name)] > 0 } {
		set TaskColor($name) yellow
	    }
	}
	return $TaskColor($name)
    }

    proc color_taskbutton {name} {
	global Pid Started Starts Now
	variable TaskButton
	if { [info exists TaskButton($name)]
	    && [winfo exists $TaskButton($name)] } { 
	    set color [taskcolor $name]
	    $TaskButton($name) config -bg $color
	}
    }

    proc check_rtsys {} {
# reread the rtsys database if necessary and update the pids
	variable RtsysFile
	variable RtsysMtime
	set mtime [file mtime $RtsysFile ]
	if { $mtime != $RtsysMtime } { 
	    set RtsysMtime $mtime

	    variable Rtsys
	    global Pid Started Starts Now
	    variable Run
	    variable OrbWidgets
	    variable OrbMeasure
	    variable Cmdline
	    variable TaskButton
	    set n [dbquery $Rtsys dbRECORD_COUNT]
	    for {set i 0} {$i<$n} {incr i} {
		set name [dbgetv $Rtsys "" $i name]
		set pid [dbgetv $Rtsys "" $i pid]
		set started [dbgetv $Rtsys "" $i last_start]
		set starts [dbgetv $Rtsys "" $i  starts]

		if { [info exists Run($name)] && $Run($name) 
		    && [info exists Cmdline($name)] } {
		    set Pid($name) $pid
		    set Started($name) $started
		    set Starts($name) $starts
		    foreach widget $OrbWidgets($name) {
			gauge::config $widget \
			    -textvariable Clients${pid}($OrbMeasure($widget))
		    }
		} 

		color_taskbutton $name
	    }
	}
    }

    proc restart {name} { 
# restart task $name
# This is done by just adding or subtracting a space following the initial command
# name
	variable TaskButton

	global Pid
	set variable Restart
	global $variable
	set $variable 1
	set w .wait
	after 10 "Procwin::wait_for_stop $name $variable $Pid($name) $w.message"
	after 10 "Procwin::dialogMessage $w Restart {stopping $name} $variable"

	pfread rtexec
	set processes [pfgetlist @rtexec#Processes]
	foreach line $processes {
	    set lname [lkey $line]
	    if { [string compare $lname $name] == 0 } { 
		set exec [lval $line]
		break 
	    }
	}

	if { [string length $exec] < 1 } { 
	    bgerror "can't find execution line for $name"
	    return
	}

	# check for extra space after command name
	set m [string first { } $exec]

	if { $m < 0 } { 
	    global Pid
	    if { [info exists Pid($name)] && $Pid($name) > 0 } {
		exec kill $Pid($name)
	    }
	    set new $exec
	} else { 
	    set n [expr $m+1]
	    set c [string range $exec $n $n]
	    # delete extra space, or add an extra space
	    if { [string compare $c { }] == 0 } { 
		incr n
		set new [string range $exec 0 $m][string range $exec $n end]
	    } else { 
		set new [string range $exec 0 $m][string range $exec $m end]
	    }
	}
	set w .notMapped
	catch [destroy $w]
	frame $w
	modifyExec $w $name $new
    }

    proc setupTaskMenu {menu name} {
	global Pid Started Starts Now
	$menu delete 0 end

	variable Cmdline
	set cmdline $Cmdline($name)

	variable Run
	switch -- $name {
	    rtm   { 
		    $menu add command \
			-state disabled \
			-label {Use File->Quit to stop}
		    $menu add command \
			-label "" \
			-state disabled 
	    }
	    rtexec {
		if { [info exists Started($name)] } {
		    $menu add command \
			-label "started [strtime $Started($name)]" \
			-state disabled
		}
		$menu add command \
		    -label "View log" \
		    -command "tailFile {$name log} $name logs/$name destroy"
		set pf rtexec.pf
		$menu add command \
		    -label "Edit parameter file $pf" \
		    -command "edit_file $pf"

		$menu add separator
		if { [updown] == "up" } { 
		    $menu add command \
			-label "Restart" \
			-command "exec rtexec -f >& logs/restart"
		} else { 
		    $menu add command \
			-label "Retry shutdown" \
			-command "exec rtexec -k < /dev/null >& logs/restart"
		}
	    }
	    default { 
		$menu add command \
		    -label $cmdline \
		    -foreground white \
		    -background #c3c3c3 \
		    -command "editExec $name"

		if { [info exists Started($name)] } {
		    if { $Starts($name) > 1 } { 
			global Now
			$menu add command \
			    -label "started $Starts($name) times, most recently [epoch2str $Started($name) {%Y-%j %H:%M}] ([string trim [strtdelta [expr $Now-$Started($name)]]] ago)" \
			    -background yellow \
			    -state disabled
		    } else { 
			$menu add command \
			    -foreground blue \
			    -label "started [strtime $Started($name)]" \
			    -state disabled
		    }
		}

		if { [file exists logs/$name] } { 
		    $menu add command \
			-label "View log" \
			-command "tailFile {$name log} $name logs/$name destroy"
		}

		variable Parameter_files
		set pf no#pf
		if { [info exists Parameter_files($name)] } { 
		    set pf pf/$Parameter_files($name)
		}

		if { ! [file exists $pf] && [regsub -- {.*-pf\s+(\S+).*} $cmdline \\1 pf] } {
		    set pf pf/$pf.pf
		}

		if { ! [file exists $pf] && [regsub -- {.*-p\s+(\S+).*} $cmdline \\1 pf] } {
		    set pf pf/$pf.pf
		}

		if { ! [file exists $pf] } {
		    set pf pf/$name.pf
		}
		
		if { ! [file exists $pf] } {
		    set pf pf/[lkey $cmdline].pf
		}

		if { [file exists $pf] } {
		    $menu add command \
			-label "Edit parameter file $pf" \
			-command "edit_file $pf"
		}

		if { [regsub -- {.*-state\s+(\S+).*} $cmdline \\1 state] } {
		    ;
		} elseif { [regsub -- {.*-S\s+(\S+).*} $cmdline \\1 state] } {
		    ;
		} else { 
		    set state state/[lkey $cmdline]
		}
		if { [file exists $state] } {
		    if { ! $Run($name) || ! ([info exists Pid($name)] && [pid_exists $Pid($name)]) } { 
			$menu add command \
			    -label "Delete state file $state" \
			    -command "clear_state $state"
		    }
		}

		$menu add separator
		if { $Run($name) } {
		    $menu add command \
			-command "Procwin::stopStartProc $name" \
			-label {Don't Run} 
		    global Pid
		    if { [info exists Pid($name)] && [updown] == "up" } { 
			$menu add command \
			    -label "Restart" \
			    -command "Procwin::restart $name" 
		    }
		} else { 
		    $menu add command \
			-command "Procwin::stopStartProc $name" \
			-label {Run} 
		}

		$menu add command \
		    -label "Delete this task" \
		    -command "modifyRun $name disable" 

		variable TaskColor
		if { [info exists TaskColor($name)] && "$TaskColor($name)" == "yellow" } { 
		    $menu add command \
			-label "Reset yellow" \
			-command "reset_yellow $name" 
		}
	    }
	}
	setTaskButton $name
    }

    proc setTaskButton {name} {
	variable Run
	variable TaskButton
	set w $TaskButton($name) 
	if { $Run($name) } { 
	    $w config -font Task
	} else { 
	    $w config -font TaskOff
	}
    }

    proc updateProcwin {} {
# perform all updates for process window
	global Pid
	global Color
	global Now
	global Started Starts

	check_rtexec
	check_rtsys

	variable TaskButton
	variable Dim
	variable Cmdline
	variable Run
	foreach name [array names Cmdline] {
	    if { [info exists Run($name)] && $Run($name) } { 

		if { ! [info exists Pid($name)] } { 
		    set Pid($name) ""
		} 
		set pid $Pid($name)
		if { $pid != "" } {
		    set statproc [statproc $Pid($name) T_$name] 
		    if { $pid < 0 || $statproc } {
			if { ! $Dim($name) } {
			    color_taskbutton $name
			    seeTask $name
			    dimActivity $name
			    set Dim($name) 1
			}
		    } else {
			if { $Dim($name) } {
			    color_taskbutton $name
			    normalActivity $name
			    set Dim($name) 0
			}
		    } 
		}
	    }
	}
	catch resize
    }

    proc seeTask {name} {
	variable TaskButton
	variable Frame
	set y [winfo y $TaskButton($name)]
	set t [winfo height $Frame.c.f]
	set range [$Frame.c yview]
	set lo [lindex $range 0]
	set hi [lindex $range 1]
	set task [expr $y.0/$t.0]
	if { $lo > $task || $hi < $task } { 
	    $Frame.c yview moveto $task
	}
    }

    proc setGauge {column name} { 
	set ::Procwin::Stat($column) $name
	if { [info exists Procwin::Gauges($column)] } { 
	    foreach widget $::Procwin::Gauges($column) { 
		if { [winfo exists $widget] } {
		    gauge::config $widget \
			-range 1 \
			-textvariable $::Procwin::Variable($widget)($name)
		} 
	    }
	}
    }

    proc setLabel {column name} { 
	# variable Procwin::Labels
	# variable Procwin::Variable
	# variable Procwin::Stat
	set ::Procwin::Stat($column) $name
	if { [info exists Procwin::Labels($column)] } { 
	    foreach widget $::Procwin::Labels($column) {
		if { [winfo exists $widget] } {
		    $widget config -textvariable $::Procwin::Variable($widget)($name)
		} 
	    }
	}
    }
    
    proc tasklabels {w} { 
# create the labels/selectorButtons for the process window
	label $w.name -text Task 
	label $w.pid -text Pid

	variable ProcStats
	variable Stat

	set Stat(l1) cpu
	selectorButton::create $w.lbl_l1 \
	    -options $ProcStats \
	    -command "Procwin::setLabel l1"

	set Stat(g1) cpu
	selectorButton::create $w.lbl_g1 \
	    -options  $ProcStats \
	    -command "Procwin::setGauge g1"

	set Stat(l2) mem
	selectorButton::create $w.lbl_l2 \
	    -options $ProcStats \
	    -command "Procwin::setLabel l2"

	set Stat(g2) mem
	selectorButton::create $w.lbl_g2 \
	    -options  $ProcStats \
	    -command "Procwin::setGauge g2"
	
	label $w.lbl_ipktrate -text "To Orb"
	label $w.lbl_opktrate -text "From Orb"
	label $w.lbl_latency -text "Latency"
	set row 0
	set col -1
	grid $w.name -row $row -column [incr col] -sticky w
	grid $w.pid -row $row -column [incr col] -sticky ew
	grid $w.lbl_l1 -row $row -column [incr col] -sticky ew
	grid $w.lbl_g1 -row $row -column [incr col] -sticky ew
	grid $w.lbl_l2 -row $row -column [incr col] -sticky ew
	grid $w.lbl_g2 -row $row -column [incr col] -sticky ew
	grid $w.lbl_ipktrate -row $row -column [incr col] -sticky ew
	grid $w.lbl_opktrate -row $row -column [incr col] -sticky ew
	grid $w.lbl_latency -row $row -column [incr col] -sticky ew
    }

    proc stopStartProc {name} { 
# toggle the Run array for $name 
	variable Run

	set flag Pid($name)
	if { $Run($name) } { 
	    modifyRun $name no
	    set title "Please wait"
	    set msg "Stopping $name"
	} else { 
	    modifyRun $name yes
	    set title "Please wait"
	    set msg "Starting $name"
	    set Pid($name) ""
	}
	set w .wait
	if { [updown] == "up" } { 
	    after 10 "Procwin::dialogMessage $w \"$title\" \"$msg\" $flag"
	}
    }

    proc taskButton {w run name cmdline} {
# create a taskButton named $w for $name with specified cmdline
	variable TaskButton
	set TaskButton($name) $w
	global ResetTime
	set ResetTime($name) 0

	menubutton $w \
	    -width 15 \
	    -text $name \
	    -menu $w.menu \
	    -anchor w 
	    
	menu $w.menu \
	    -tearoff no \
	    -disabledforeground blue \
	    -postcommand "Procwin::setupTaskMenu $w.menu $name"

	set_balloon $w $cmdline

	grid $w -sticky w
	setTaskButton $name
    }

    proc createProcwin {w} { 

	frame $w
	variable Frame
	set Frame $w
	variable HasScrollbar
	set HasScrollbar ""
	scrollbar $w.vscroll -command "$w.c yview"
	canvas $w.c \
	    -yscrollcommand "$w.vscroll set"
	grid $w.vscroll -row 0 -column 0 -ipadx 1 -sticky ns
	grid $w.c	-row 0 -column 1 -ipadx 2 -sticky nsew 
	set f $w.c.f
	frame $f
	$w.c create window 0 0 -window $f -anchor nw

	variable Root
	set Root $f

	global Pf
	global YellowTime 
	set YellowTime [str2epoch [pfget $Pf yellow_time]]

	variable ProcStats
	set ProcStats [pfgetlist @$Pf#ProcStats]
	variable ShowRtm
	set ShowRtm [pfget $Pf show_rtm]

	tasklabels $f

	# update idletasks
	variable HeaderHeight
	set HeaderHeight [winfo reqheight $f]

	variable RtexecMtime
	set RtexecMtime 0
	check_rtexec

	set RtsysMtime 0
	check_rtsys

	selectorButton::choose $f.lbl_l1 cpu
	selectorButton::choose $f.lbl_g1 cpu
	selectorButton::choose $f.lbl_l2 rss
	selectorButton::choose $f.lbl_g2 rss

	Procwin::updateProcwin
	
	global ProcessWindowCreated 
	set ProcessWindowCreated 1
	catch resize
    }

    proc resize {} { 
	# update idletasks
	global ProcessWindowCreated 
	if { $ProcessWindowCreated == 0 } { 
	    return
	}
	variable Root
	set y2 [winfo reqheight $Root]
	variable TaskButton
	set n [llength [array names TaskButton]]
	variable HeaderHeight
	set per [expr ($y2-$HeaderHeight)/$n ]
	global Pf
	set process_window_max_size [pfget $Pf process_window_max_size]
	set vrootheight [winfo vrootheight .] 
	set maxheight 200
	catch { set maxheight [expr $vrootheight * $process_window_max_size] }
	set x [winfo reqwidth $Root]
	variable Frame
	variable HasScrollbar
	if { $y2 > $maxheight } { 
		grid $Frame.vscroll -row 0 -column 0 -ipadx 1 -sticky ns
		grid $Frame.c	-row 0 -column 1 -ipadx 2 -sticky nsew 
		set nmax [expr int($maxheight/$per)]
		set y [expr $nmax*$per + $HeaderHeight + 2]
		$Frame.c configure -scrollregion \
		    [list 0 0 [winfo reqwidth $Root] $y2] \
		    -width $x -height $y
		set HasScrollbar $Frame.vscroll 
	} else {
		grid forget $Frame.vscroll
		$Frame.c configure -width $x -height $y2 
		$Frame.c yview moveto 0
		set HasScrollbar "" 
	}
    }

    proc destroyProcwin {} { 
	global ProcessWindowCreated 
	set ProcessWindowCreated 0

	variable Gauges
	foreach column [array names Gauges] { 
	    foreach gauge $Gauges($column) {
		catch {gauge::destroy $gauge}
	    }
	}
	catch {unset -nocomplain Gauges}

	variable Root
	set w $Root
	destroy $w

	variable TaskButton
	catch {unset -nocomplain TaskButton}
	variable Activity
	catch {unset -nocomplain Activity}
	variable Dim
	catch {unset -nocomplain Dim}
	global Pid
	catch {unset -nocomplain Pid}
	variable Run
	catch {unset -nocomplain Run}
	variable OrbWidgets
	catch {unset -nocomplain OrbWidgets}
    }

    namespace export createProcwin destroyProcwin
}

proc reset_yellow {name} { 
    global ResetTime Now
    set ResetTime($name) $Now 
    Procwin::color_taskbutton $name
}
